package me.regexp;

import java.util.*;

public final class RE {
	
    /**
     * Specifies normal, case-sensitive matching behaviour.
     */
    public static final int MATCH_NORMAL = 0x0000;
    /**
     * Flag to indicate that matching should be case-independent (folded)
     */
    public static final int MATCH_CASEINDEPENDENT = 0x0001;
    /**
     * Newlines should match as BOL/EOL (^ and $)
     */
    public static final int MATCH_MULTILINE = 0x0002;
    /**
     * Consider all input a single body of text - newlines are matched by .
     */
    public static final int MATCH_SINGLELINE = 0x0004;
    /************************************************
     *                                              *
     * The format of a node in a program is:        *
     *                                              *
     * [ OPCODE ] [ OPDATA ] [ OPNEXT ] [ OPERAND ] *
     *                                              *
     * char OPCODE - instruction                    *
     * char OPDATA - modifying data                 *
     * char OPNEXT - next node (relative offset)    *
     *                                              *
     ************************************************/

                 //   Opcode              Char       Opdata/Operand  Meaning
                 //   ----------          ---------- --------------- --------------------------------------------------
    static final char OP_END              = 'E';  //                 end of program
    static final char OP_BOL              = '^';  //                 match only if at beginning of line
    static final char OP_EOL              = '$';  //                 match only if at end of line
    static final char OP_ANY              = '.';  //                 match any single character except newline
    static final char OP_ANYOF            = '[';  // count/ranges    match any char in the list of ranges
    static final char OP_BRANCH           = '|';  // node            match this alternative or the next one
    static final char OP_ATOM             = 'A';  // length/string   length of string followed by string itself
    static final char OP_STAR             = '*';  // node            kleene closure
    static final char OP_PLUS             = '+';  // node            positive closure
    static final char OP_MAYBE            = '?';  // node            optional closure
    static final char OP_ESCAPE           = '\\'; // escape          special escape code char class (escape is E_* code)
    static final char OP_OPEN             = '(';  // number          nth opening paren
    static final char OP_OPEN_CLUSTER     = '<';  //                 opening cluster
    static final char OP_CLOSE            = ')';  // number          nth closing paren
    static final char OP_CLOSE_CLUSTER    = '>';  //                 closing cluster
    static final char OP_BACKREF          = '#';  // number          reference nth already matched parenthesized string
    static final char OP_GOTO             = 'G';  //                 nothing but a (back-)pointer
    static final char OP_NOTHING          = 'N';  //                 match null string such as in '(a|)'
    static final char OP_CONTINUE         = 'C';  //                 continue to the following command (ignore next)
    static final char OP_RELUCTANTSTAR    = '8';  // none/expr       reluctant '*' (mnemonic for char is unshifted '*')
    static final char OP_RELUCTANTPLUS    = '=';  // none/expr       reluctant '+' (mnemonic for char is unshifted '+')
    static final char OP_RELUCTANTMAYBE   = '/';  // none/expr       reluctant '?' (mnemonic for char is unshifted '?')
    static final char OP_POSIXCLASS       = 'P';  // classid         one of the posix character classes

    // Escape codes
    static final char E_ALNUM             = 'w';  // Alphanumeric
    static final char E_NALNUM            = 'W';  // Non-alphanumeric
    static final char E_BOUND             = 'b';  // Word boundary
    static final char E_NBOUND            = 'B';  // Non-word boundary
    static final char E_SPACE             = 's';  // Whitespace
    static final char E_NSPACE            = 'S';  // Non-whitespace
    static final char E_DIGIT             = 'd';  // Digit
    static final char E_NDIGIT            = 'D';  // Non-digit

    // Posix character classes
    static final char POSIX_CLASS_ALNUM   = 'w';  // Alphanumerics
    static final char POSIX_CLASS_ALPHA   = 'a';  // Alphabetics
    static final char POSIX_CLASS_BLANK   = 'b';  // Blanks
    static final char POSIX_CLASS_CNTRL   = 'c';  // Control characters
    static final char POSIX_CLASS_DIGIT   = 'd';  // Digits
    static final char POSIX_CLASS_GRAPH   = 'g';  // Graphic characters
    static final char POSIX_CLASS_LOWER   = 'l';  // Lowercase characters
    static final char POSIX_CLASS_PRINT   = 'p';  // Printable characters
    static final char POSIX_CLASS_PUNCT   = '!';  // Punctuation
    static final char POSIX_CLASS_SPACE   = 's';  // Spaces
    static final char POSIX_CLASS_UPPER   = 'u';  // Uppercase characters
    static final char POSIX_CLASS_XDIGIT  = 'x';  // Hexadecimal digits
    static final char POSIX_CLASS_JSTART  = 'j';  // Java identifier start
    static final char POSIX_CLASS_JPART   = 'k';  // Java identifier part
    
    // Limits
    static final int maxNode  = 65536;            // Maximum number of nodes in a program
    static final int MAX_PAREN = 16;              // Number of paren pairs (only 9 can be backrefs)

    // Node layout constants
    static final int offsetOpcode = 0;            // Opcode offset (first character)
    static final int offsetOpdata = 1;            // Opdata offset (second char)
    static final int offsetNext   = 2;            // Next index offset (third char)
    static final int nodeSize     = 3;            // Node size (in chars)

	public Class45 aClass45_421;
	
	private REProgram program;
	public transient CharacterIterator search;
	private int matchFlags;
	
	public transient int parenCount;
	private transient int start0;
	private transient int end0;
	private transient int start1;
	private transient int end1;
	private transient int start2;
	private transient int end2;
	private transient int[] startn;
	private transient int[] end;
	private transient int[] startBackref;
	private transient int[] endBackref;

	public RE(final String s, final byte b) {
		this(s);
	}

	private RE(final String s) {
		this(new RECompiler().compile(s));
	}

	private RE(final REProgram program) {
		super();
		this.matchFlags = 16;
		this.program = program;
		if (program != null && program.maxParens != -1) {
			this.matchFlags = program.maxParens;
			return;
		}
		this.matchFlags = 16;
	}

	public RE() {
		this((REProgram) null);
	}
	
	/**
     * Returns the start index of a given paren level.
     *
     * @param which Nesting level of subexpression
     * @return String index
     */
	public final int getParenStart(final int n) {
		if (n >= this.parenCount) {
			return -1;
		}
		switch (n) {
		case 0:
			return this.start0;
		case 1:
			return this.start1;
		case 2:
			return this.start2;
		default:
			if (this.startn == null) {
				this.allocParens();
			}
			return this.startn[n];
		}
	}

    /**
     * Returns the end index of a given paren level.
     *
     * @param which Nesting level of subexpression
     * @return String index
     */
	public final int getParenEnd(final int n) {
		if (n >= this.parenCount) {
			return -1;
		}
		switch (n) {
		case 0:
			return this.end0;
		case 1:
			return this.end1;
		case 2:
			return this.end2;
		default:
			if (this.end == null) {
				this.allocParens();
			}
			return this.end[n];
		}
	}

    /**
     * Sets the start of a paren level
     *
     * @param which Which paren level
     * @param i Index in input array
     */
	private void setParenStart(final int n, final int anInt434) {
		if (n < this.parenCount) {
			switch (n) {
			case 0:
				this.start0 = anInt434;
			case 1:
				this.start1 = anInt434;
			case 2:
				this.start2 = anInt434;
			default:
				if (this.startn == null) {
					this.allocParens();
				}
				this.startn[n] = anInt434;
				break;
			}
		}
	}
	
    /**
     * Sets the end of a paren level
     *
     * @param which Which paren level
     * @param i Index in input array
     */
	private void setParenEnd(final int n, final int anInt435) {
		if (n < this.parenCount) {
			switch (n) {
			case 0:
				this.end0 = anInt435;
			case 1:
				this.end1 = anInt435;
			case 2:
				this.end2 = anInt435;
			default:
				if (this.end == null) {
					this.allocParens();
				}
				this.end[n] = anInt435;
				break;
			}
		}
	}

	private static void internalError(final String s) {
		throw new Error("RE internal error: " + s);
	}

	private void allocParens() {
		this.startn = new int[this.matchFlags];
		this.end = new int[this.matchFlags];
		for (int i = 0; i < this.matchFlags; ++i) {
			this.startn[i] = -1;
			this.end[i] = -1;
		}
	}

    protected int matchNodes(int firstNode, int lastNode, int idxStart) {
        // Our current place in the string
        int idx = idxStart;

        // Loop while node is valid
        int next, opcode, opdata;
        int idxNew;
        char[] instruction = program.instruction;
        for (int node = firstNode; node < lastNode;) {
            opcode = instruction[node /* + offsetOpcode */];
            next = node + (short) instruction[node + offsetNext];
            opdata = instruction[node + offsetOpdata];

            switch (opcode) {
                case OP_MAYBE:
                case OP_STAR: {
                    // Try to match the following subexpr. If it matches:
                    //   MAYBE:  Continues matching rest of the expression
                    //    STAR:  Points back here to repeat subexpr matching
                    if ((idxNew = matchNodes(node + nodeSize, maxNode, idx)) != -1) {
                        return idxNew;
                    }

                    // If failed, just continue with the rest of expression
                    break;
                }

                case OP_PLUS: {
                    // Try to match the subexpr again (and again (and ...
                    if ((idxNew = matchNodes(next, maxNode, idx)) != -1) {
                        return idxNew;
                    }

                    // If failed, just continue with the rest of expression
                    // Rest is located at the next pointer of the next instruction
                    // (which must be OP_CONTINUE)
                    node = next + (short) instruction[next + offsetNext];
                    continue;
                }

                case OP_RELUCTANTMAYBE:
                case OP_RELUCTANTSTAR: {
                    // Try to match the rest without using the reluctant subexpr
                    if ((idxNew = matchNodes(next, maxNode, idx)) != -1) {
                        return idxNew;
                    }

                    // Try reluctant subexpr. If it matches:
                    //   RELUCTANTMAYBE: Continues matching rest of the expression
                    //    RELUCTANTSTAR: Points back here to repeat reluctant star matching
                    return matchNodes(node + nodeSize, next, idx);
                }

                case OP_RELUCTANTPLUS: {
                    // Continue matching the rest without using the reluctant subexpr
                    if ((idxNew = matchNodes(next + (short) instruction[next + offsetNext], maxNode, idx)) != -1) {
                        return idxNew;
                    }

                    // Try to match subexpression again
                    break;
                }

                case OP_OPEN:

                    // Match subexpression
                    if ((program.flags & REProgram.OPT_HASBACKREFS) != 0) {
                        startBackref[opdata] = idx;
                    }
                    if ((idxNew = matchNodes(next, maxNode, idx)) != -1) {
                        // Increase valid paren count
                        if (opdata >= parenCount) {
                            parenCount = opdata + 1;
                        }

                        // Don't set paren if already set later on
                        if (getParenStart(opdata) == -1) {
                            setParenStart(opdata, idx);
                        }
                    }
                    return idxNew;

                case OP_CLOSE:

                    // Done matching subexpression
                    if ((program.flags & REProgram.OPT_HASBACKREFS) != 0) {
                        endBackref[opdata] = idx;
                    }
                    if ((idxNew = matchNodes(next, maxNode, idx)) != -1) {
                        // Increase valid paren count
                        if (opdata >= parenCount) {
                            parenCount = opdata + 1;
                        }

                        // Don't set paren if already set later on
                        if (getParenEnd(opdata) == -1) {
                            setParenEnd(opdata, idx);
                        }
                    }
                    return idxNew;

                case OP_BACKREF:
                     {
                        // Get the start and end of the backref
                        int s = startBackref[opdata];
                        int e = endBackref[opdata];

                        // We don't know the backref yet
                        if (s == -1 || e == -1) {
                            return -1;
                        }

                        // The backref is empty size
                        if (s == e) {
                            break;
                        }

                        // Get the length of the backref
                        int l = e - s;

                        // If there's not enough input left, give up.
                        if (search.isEnd(idx + l - 1)) {
                            return -1;
                        }

                        // Case fold the backref?
                        final boolean caseFold =
                                ((matchFlags & MATCH_CASEINDEPENDENT) != 0);
                        // Compare backref to input
                        for (int i = 0; i < l; i++) {
                            if (compareChars(search.charAt(idx++), search.charAt(s + i), caseFold) != 0) {
                                return -1;
                            }
                        }
                    }
                    break;

                case OP_BOL:

                    // Fail if we're not at the start of the string
                    if (idx != 0) {
                        // If we're multiline matching, we could still be at the start of a line
                        if ((matchFlags & MATCH_MULTILINE) == MATCH_MULTILINE) {
                            // Continue if at the start of a line
                            if (isNewline(idx - 1)) {
                                break;
                            }
                        }
                        return -1;
                    }
                    break;

                case OP_EOL:

                    // If we're not at the end of string
                    if (!search.isEnd(0) && !search.isEnd(idx)) {
                        // If we're multi-line matching
                        if ((matchFlags & MATCH_MULTILINE) == MATCH_MULTILINE) {
                            // Continue if we're at the end of a line
                            if (isNewline(idx)) {
                                break;
                            }
                        }
                        return -1;
                    }
                    break;

                case OP_ESCAPE:

                    // Which escape?
                    switch (opdata) {
                        // Word boundary match
                        case E_NBOUND:
                        case E_BOUND:
                             {
                                char cLast = ((idx == 0) ? '\n' : search.charAt(idx - 1));
                                char cNext = ((search.isEnd(idx)) ? '\n' : search.charAt(idx));
                                if ((RECharacter.isLetterOrDigit(cLast) == RECharacter.isLetterOrDigit(cNext)) == (opdata == E_BOUND)) {
                                    return -1;
                                }
                            }
                            break;

                        // Alpha-numeric, digit, space, javaLetter, javaLetterOrDigit
                        case E_ALNUM:
                        case E_NALNUM:
                        case E_DIGIT:
                        case E_NDIGIT:
                        case E_SPACE:
                        case E_NSPACE:

                            // Give up if out of input
                            if (search.isEnd(idx)) {
                                return -1;
                            }

                            char c = search.charAt(idx);

                            // Switch on escape
                            switch (opdata) {
                                case E_ALNUM:
                                case E_NALNUM:
                                    if (!((RECharacter.isLetterOrDigit(c) || c == '_') == (opdata == E_ALNUM))) {
                                        return -1;
                                    }
                                    break;

                                case E_DIGIT:
                                case E_NDIGIT:
                                    if (!(RECharacter.isDigit(c) == (opdata == E_DIGIT))) {
                                        return -1;
                                    }
                                    break;

                                case E_SPACE:
                                case E_NSPACE:
                                    if (!(RECharacter.isWhitespace(c) == (opdata == E_SPACE))) {
                                        return -1;
                                    }
                                    break;
                            }
                            idx++;
                            break;

                        default:
                            internalError("Unrecognized escape '" + opdata + "'");
                    }
                    break;

                case OP_ANY:

                    if ((matchFlags & MATCH_SINGLELINE) == MATCH_SINGLELINE) {
                        // Match anything
                        if (search.isEnd(idx)) {
                            return -1;
                        }
                    } else {
                        // Match anything but a newline
                        if (search.isEnd(idx) || isNewline(idx)) {
                            return -1;
                        }
                    }
                    idx++;
                    break;

                case OP_ATOM:
                     {
                        // Match an atom value
                        if (search.isEnd(idx)) {
                            return -1;
                        }

                        // Get length of atom and starting index
                        // int lenAtom = opdata;
                        int startAtom = node + nodeSize;

                        // Give up if not enough input remains to have a match
                        if (search.isEnd(opdata + idx - 1)) {
                            return -1;
                        }

                        // Match atom differently depending on casefolding flag
                        final boolean caseFold =
                                ((matchFlags & MATCH_CASEINDEPENDENT) != 0);

                        for (int i = 0; i < opdata; i++) {
                            if (compareChars(search.charAt(idx++), instruction[startAtom + i], caseFold) != 0) {
                                return -1;
                            }
                        }
                    }
                    break;

                case OP_POSIXCLASS:
                     {
                        // Out of input?
                        if (search.isEnd(idx)) {
                            return -1;
                        }

                        switch (opdata) {
                            case POSIX_CLASS_ALNUM:
                                if (!RECharacter.isLetterOrDigit(search.charAt(idx))) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_ALPHA:
                                if (!RECharacter.isLetter(search.charAt(idx))) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_DIGIT:
                                if (!RECharacter.isDigit(search.charAt(idx))) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_BLANK: // JWL - bugbug: is this right??
                                if (!RECharacter.isSpaceChar(search.charAt(idx))) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_SPACE:
                                if (!RECharacter.isWhitespace(search.charAt(idx))) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_CNTRL:
                                if (RECharacter.getType(search.charAt(idx)) != RECharacter.CONTROL) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_GRAPH: // JWL - bugbug???
                                switch (RECharacter.getType(search.charAt(idx))) {
                                    case RECharacter.MATH_SYMBOL:
                                    case RECharacter.CURRENCY_SYMBOL:
                                    case RECharacter.MODIFIER_SYMBOL:
                                    case RECharacter.OTHER_SYMBOL:
                                        break;

                                    default:
                                        return -1;
                                }
                                break;

                            case POSIX_CLASS_LOWER:
                                if (RECharacter.getType(search.charAt(idx)) != RECharacter.LOWERCASE_LETTER) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_UPPER:
                                if (RECharacter.getType(search.charAt(idx)) != RECharacter.UPPERCASE_LETTER) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_PRINT:
                                if (RECharacter.getType(search.charAt(idx)) == RECharacter.CONTROL) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_PUNCT:
                                 {
                                    int type = RECharacter.getType(search.charAt(idx));
                                    switch (type) {
                                        case RECharacter.DASH_PUNCTUATION:
                                        case RECharacter.START_PUNCTUATION:
                                        case RECharacter.END_PUNCTUATION:
                                        case RECharacter.CONNECTOR_PUNCTUATION:
                                        case RECharacter.OTHER_PUNCTUATION:
                                            break;

                                        default:
                                            return -1;
                                    }
                                }
                                break;

                            case POSIX_CLASS_XDIGIT: // JWL - bugbug??
                                 {
                                    boolean isXDigit = ((search.charAt(idx) >= '0' && search.charAt(idx) <= '9') ||
                                            (search.charAt(idx) >= 'a' && search.charAt(idx) <= 'f') ||
                                            (search.charAt(idx) >= 'A' && search.charAt(idx) <= 'F'));
                                    if (!isXDigit) {
                                        return -1;
                                    }
                                }
                                break;

                            case POSIX_CLASS_JSTART:
                                if (!RECharacter.isJavaIdentifierStart(search.charAt(idx))) {
                                    return -1;
                                }
                                break;

                            case POSIX_CLASS_JPART:
                                if (!RECharacter.isJavaIdentifierPart(search.charAt(idx))) {
                                    return -1;
                                }
                                break;

                            default:
                                internalError("Bad posix class");
                                break;
                        }

                        // Matched.
                        idx++;
                    }
                    break;

                case OP_ANYOF:
                     {
                        // Out of input?
                        if (search.isEnd(idx)) {
                            return -1;
                        }

                        // Get character to match against character class and maybe casefold
                        char c = search.charAt(idx);
                        boolean caseFold = (matchFlags & MATCH_CASEINDEPENDENT) != 0;
                        // Loop through character class checking our match character
                        int idxRange = node + nodeSize;
                        int idxEnd = idxRange + (opdata * 2);
                        boolean match = false;
                        for (int i = idxRange; !match && i < idxEnd;) {
                            // Get start, end and match characters
                            char s = instruction[i++];
                            char e = instruction[i++];

                            match = ((compareChars(c, s, caseFold) >= 0) && (compareChars(c, e, caseFold) <= 0));
                        }

                        // Fail if we didn't match the character class
                        if (!match) {
                            return -1;
                        }
                        idx++;
                    }
                    break;

                case OP_BRANCH: {
                    // Check for choices
                    // FIXME Dead code - only reason to keep is backward compat with pre-compiled exprs. Remove?
                    if (instruction[next /* + offsetOpcode */] != OP_BRANCH) {
                        // If there aren't any other choices, just evaluate this branch.
                        node += nodeSize;
                        continue;
                    }

                    // Try all available branches
                    int nextBranch;
                    do {
                        // Try matching the branch against the string
                        if ((idxNew = matchNodes(node + nodeSize, maxNode, idx)) != -1) {
                            return idxNew;
                        }

                        // Go to next branch (if any)
                        nextBranch = (short) instruction[node + offsetNext];
                        node += nextBranch;
                    } while (nextBranch != 0 && (instruction[node /* + offsetOpcode */] == OP_BRANCH));

                    // Failed to match any branch!
                    return -1;
                }

                case OP_OPEN_CLUSTER:
                case OP_CLOSE_CLUSTER:
                // starting or ending the matching of a subexpression which has no backref.

                case OP_NOTHING:
                case OP_GOTO:

                    // Just advance to the next node without doing anything
                    break;

                case OP_CONTINUE:

                    // Advance to the following node
                    node += nodeSize;
                    continue;

                case OP_END:

                    // Match has succeeded!
                    setParenEnd(0, idx);
                    return idx;

                default:

                    // Corrupt program
                    internalError("Invalid opcode '" + opcode + "'");
            }

            // Advance to the next node in the program
            node = next;
        }

        // We "should" never end up here
        internalError("Corrupt program");
        return -1;
    }

	private boolean matchAt(int method376) {
		this.start0 = -1;
		this.end0 = -1;
		this.start1 = -1;
		this.end1 = -1;
		this.start2 = -1;
		this.end2 = -1;
		this.startn = null;
		this.end = null;
		this.parenCount = 1;
		this.setParenStart(0, method376);
		if ((this.program.flags & 0x1) != 0x0) {
			this.startBackref = new int[this.matchFlags];
			this.endBackref = new int[this.matchFlags];
		}
		if ((method376 = this.matchNodes(0, 65536, method376)) != -1) {
			this.setParenEnd(0, method376);
			return true;
		}
		this.parenCount = 0;
		return false;
	}

	public final boolean match(final String s, final int n) {
		final CharacterIterator class34 = new CharacterIterator(s);
		int n2 = n;
		final CharacterIterator aClass34_423 = class34;
		if (this.program == null) {
			internalError("No RE program to run!");
		}
		this.search = aClass34_423;
		if ((this.program.flags & 0x2) == 0x2) {
			return n2 == 0 && this.matchAt(n2);
		}
		if (this.program.prefix == null) {
			while (!aClass34_423.isEnd(n2 - 1)) {
				if (this.matchAt(n2)) {
					return true;
				}
				++n2;
			}
			return false;
		}
		for (char[] aCharArray439 = this.program.prefix; !aClass34_423
				.isEnd(n2 + aCharArray439.length - 1); ++n2) {
			int n3 = n2;
			int n4 = 0;
			while (aClass34_423.charAt(n3++) - aCharArray439[n4++] == '\0' && n4 < aCharArray439.length) {
			}
			if (n4 == aCharArray439.length && this.matchAt(n2)) {
				return true;
			}
		}
		return false;
	}

	public final Vector splitV(final String s) {
		final Vector vector = new Vector();
		int n;
		int method372;
		for (n = 0; n < s.length() && this.match(s, n); n = method372) {
			final int method371 = this.getParenStart(0);
			if ((method372 = this.getParenEnd(0)) == n) {
				vector.addElement(s.substring(n, method371 + 1));
				++method372;
			} else {
				vector.addElement(s.substring(n, method371));
			}
		}
		vector.addElement(s.substring(n));
		return vector;
	}

	public final String[] split(final String s) {
		final Vector method379;
		final String[] array = new String[(method379 = this.splitV(s)).size()];
		method379.copyInto(array);
		return array;
	}

    /**
     * Matches the current regular expression program against a String.
     *
     * @param search String to match against
     * @return True if string matched
     */
    public boolean match(String search) {
        return match(search, 0);
    }

	/**
     * @return true if character at i-th position in the <code>search</code> string is a newline
     */
    private boolean isNewline(int i) {
        char nextChar = search.charAt(i);

        return nextChar == '\n' || nextChar == '\r' || nextChar == '\u0085' ||
                nextChar == '\u2028' || nextChar == '\u2029';
    }

    /**
     * Compares two characters.
     *
     * @param c1 first character to compare.
     * @param c2 second character to compare.
     * @param caseIndependent whether comparision is case insensitive or not.
     * @return negative, 0, or positive integer as the first character
     *         less than, equal to, or greater then the second.
     */
    private int compareChars(char c1, char c2, boolean caseIndependent) {
        if (caseIndependent) {
            c1 = RECharacter.toLowerCase(c1);
            c2 = RECharacter.toLowerCase(c2);
        }
        return ((int) c1 - (int) c2);
    }

	private static int compareChars(final int j, final int m) {
		return j - m;
	}

	public final String method373(final String s) {
		if (this.match(s, 0)) {
			return s.substring(this.getParenStart(0), this.getParenEnd(0));
		}
		return null;
	}
}
